# -=FISH Version 0.1=-
# Contact: Sarid Harper, saridski@yahoo.dk
# Written: 01.2006

# Please send any questions or comments to the address above
# Feel free to do what ever you wish with this programme and its code

# Install XML-XPath before using this programme, if running it with the 
# Perl interpreter
 
package filecheck;
	use strict;
	use XML::XPath;
	use Digest::MD5 qw(md5_hex);
	use Digest::SHA1 qw(sha1_hex);
	use configuration;

	# ----------------------------------------------	
	# PUBLIC MEMBERS
	# ----------------------------------------------	
	sub filecheck::calculate($) {
		# This is the main worker public member
		my $self = shift();
		
		print("\n-=FISH Version 0.1=-\n(F)ile\n(I)ntegrity tool by\n(S)arid\n(H)arper (sharpe)\n");
		print("\nIterating file system.. ");
		$self->_walkfs($self->{_CONFIGURATION}->top());
		print("Complete.\n");
		print("Create signature database (".$self->{_CONFIGURATION}->output().").. ");
		$self->_createdatabase();
		print("Complete.\n");
		if ((keys %{$self->{_LOG}}) > 0) {
			print("Writing error log.. ");
			$self->_writeerrorlog();
			print("Complete.\n");
		}
	}
	
	sub filecheck::comparefiles($$$) {
		# This public member is called when a user whishes to compare an older (known good)
		# signature database with the current system files
		my $self = shift();
		my $file = shift();
		my $report = shift();

		print("Reading archived signature database (".$self->{_CONFIGURATION}->sigdb().").. ");
		$self->_comparefiles($self->{_CONFIGURATION}->sigdb());			
		print("Complete.\nLooking for deleted files.. ");
		$self->_findnewfiles();
		print("Complete\nLooking for new files...");
		$self->_finddeletedfiles();			
		print("Complete.\nWriting results (".$self->{_CONFIGURATION}->report().").. ");
		$self->_createreport($self->{_CONFIGURATION}->report());	
		print("Complete.\nFile system integrity check complete.\n");
		print("\nSummary: Altered Files: ".(keys %{$self->{_ALTEREDFILES}})."\n\t New files: ".(keys %{$self->{_NEWFILES}})."\n\t Deleted files: ".(keys %{$self->{_DELETEDFILES}})."\n");
	}

	sub filecheck::NEW($$) {
		# Constructor
		my $proto = shift();
		my $class = ref($proto) || $proto;
		my $self  = {};	
		
		$self->{_CONFIGURATION} = configuration->NEW(shift());
		$self->{_TOP} = undef;	
		$self->{_XMLDB} = undef;
		$self->{_HASH} = undef;
		$self->{_FILETYPES} = ();
		$self->{_LOG} = undef;
		$self->{_CURRENTFILES} = ();
		$self->{_ALTEREDFILES} = ();
		$self->{_DBFILES} = ();	
		$self->{_NEWFILES} = ();
		$self->{_DELETEDFILES} = ();
		$self->{_COUNT} = 0;
		
		bless ($self, $class);
		return $self;
	}
	
	sub filecheck::DESTROY($) {
		# Destructor
		my $self = shift();
		
		$self->{_CONFIGURATION} = configuration->NEW(shift());
		$self->{_TOP} = undef;	
		$self->{_XMLDB} = undef;
		$self->{_HASH} = undef;
		$self->{_FILETYPES} = ();
		$self->{_LOG} = undef;
		$self->{_CURRENTFILES} = ();
		$self->{_ALTEREDFILES} = ();
		$self->{_DBFILES} = ();	
		$self->{_NEWFILES} = ();
		$self->{_DELETEDFILES} = ();
		$self->{_COUNT} = 0;
	}	
	
	# ----------------------------------------------	
	# PRIVATE MEMBERS
	# ----------------------------------------------	
	sub filecheck::_walkfs($$) { 
		# This private member iterates the file system recursively
		my $self = shift();
		my $top = shift();
		my $dir = undef;

		if (-d $top) {
			my $file = undef;
			if (opendir($dir, $top)) {
				while ($file = readdir($dir)) {
					next if ($file eq '.' || $file eq '..');
					foreach my $ext (@{$self->{_CONFIGURATION}->filetypes()}) {
						if (-f "$top/$file") {
							# Removed \. to enable the hashing of all files when no extension is given
							if ($file =~ /$ext$/i) {
								$self->{_COUNT}++;
								$self->{_CURRENTFILES}->{"$top/$file"} = $self->_computehash("$top/$file");
							}
						}
					}
					# Recursion in action. Calling itself..
					$self->_walkfs("$top/$file");
				}
			} else {
				return;
			}
		}
	}		
	
	sub filecheck::_computehash($$) {
		# Very basic stuff here..		
		# Opens file name passed to this private method
		# and calculates either a MD5 or SHA1 hash of its contents
		my $self = shift();
		my $file = shift();
		
		
		open(F_DATA, "<", $file) || $self->_log($file, $!);
		my $contents = do { local $/; <F_DATA> }; 
		close(F_DATA);
		
		if ($self->{_HASH} eq "MD5") {
			return Digest::MD5->md5_hex($contents)
		} else {
			return Digest::SHA1->sha1_hex($contents)
		}
	}	
	
	sub filecheck::_createdatabase($) {
		# Creates an XML database containing the signatures of all the included files
		# This file should be stored in a secure place, as it contains the file signatures
		my $self = shift();
		
		open(F_OUT, ">", ($self->{_CONFIGURATION}->output())) || die($!);
		
		print(F_OUT qq{<?xml version="1.0" encoding="iso-8859-1"?>\n});
		print(F_OUT qq{<signatures rootdir="}.$self->{_CONFIGURATION}->top().qq{" filetypes="}.join(":", @{$self->{_CONFIGURATION}->filetypes()}).qq{" count="$self->{_COUNT}" created="}.localtime(time()).qq{" hash="}.$self->{_CONFIGURATION}->hash().qq{">\n});

		foreach (keys %{$self->{_CURRENTFILES}}) {
			print(F_OUT qq{\t<signature hash="$self->{_CURRENTFILES}->{$_}"><![CDATA[$_]]></signature>\n});
		}
		
		print(F_OUT qq{</signatures>\n});
		close(F_OUT);
	}

	sub filecheck::_comparefiles($$) {
		# Opens the signature database (the xml file generated by the method above)
		# Iterates through it and compared the the hash value in the signature data
		# with the hash just generated, which is located in this package's private member
		# called $self->{_CURRENTFILES}
		my $self = shift();
		my $database = shift();
		
		if (-e $database) {
			my $xpath = XML::XPath->new(filename => $database) || print($!);
			my $nodeset = $xpath->find('//signatures/signature');
			if ($self->{_CURRENTFILES}) {
				foreach ($nodeset->get_nodelist) {
					my $file = $_->string_value();
					my $hash = $_->find('@hash')->string_value();
					if (exists($self->{_CURRENTFILES}->{$file})) {
						if (($self->{_CURRENTFILES}->{$file}) ne ($hash)) {
							$self->{_ALTEREDFILES}->{$file} = $hash;
						}
					}
					$self->{_DBFILES}->{$file} = $hash;
				}
			} else {					
				die("Please calculate file signatures first.\n");
			}
		} else {
			print("Failed\nThe specified file signature database does not exist.. Aborting.\n");
			exit();
		}
	}		
	
	sub filecheck::_findnewfiles($) {
		# Looks for new files
		# If a file exists on disk but not in the signature database then it is new
		# The results are stored in the private member $self->{_NEWFILES}
		my $self = shift();
		my $key = undef;

		foreach (keys %{$self->{_CURRENTFILES}}) {
			if (!(exists($self->{_DBFILES}->{$_}))) {
				$self->{_NEWFILES}->{$_} = undef;
			}
		}
	}
	
	sub filecheck::_finddeletedfiles($) {
		# Looks for deleted files
		# If a file exists in the database but not on disk then it has been deleted
		# The results are stored in the private member $self->{_DELETEDFILES}		
		my $self = shift();

		foreach (keys %{$self->{_DBFILES}}) {
			if (!(exists($self->{_CURRENTFILES}->{$_}))) {
				$self->{_DELETEDFILES}->{$_} = undef;
			}
		}
	}	
	
	sub filecheck::_createreport($$) {
		# Creates an xml report and stores the integrity scan results
		my $self = shift();
		my $file = shift();
		
		open(F_OUT, ">", $file) || die($!);
		print(F_OUT qq{<?xml version="1.0" encoding="iso-8859-1"?>\n});
		print(F_OUT qq{<changes>\n});
		
		print(F_OUT qq{\t<alteredfiles rootdir="}.$self->{_CONFIGURATION}->top().qq{" filetypes="}.join(":", @{$self->{_CONFIGURATION}->filetypes()}).qq{" count="}.(keys %{$self->{_ALTEREDFILES}}).qq{" created="}.localtime(time()).qq{" hash="}.$self->{_CONFIGURATION}->hash().qq{">\n});
		foreach (keys %{$self->{_ALTEREDFILES}}) {
			print(F_OUT qq{\t\t<alteredfile newhash="$self->{_ALTEREDFILES}->{$_}" oldhash="$self->{_CURRENTFILES}->{$_}"><![CDATA[$_]]></alteredfile>\n});
		}
		print(F_OUT qq{\t</alteredfiles>\n});
		
		print(F_OUT qq{\t<newfiles count="}.(keys %{$self->{_NEWFILES}}).qq{">\n});
		foreach (keys %{$self->{_NEWFILES}}) {
			print(F_OUT qq{\t\t<newfile><![CDATA[$_]]></newfile>\n});
		}		
		print(F_OUT qq{\t</newfiles>\n});
		
		print(F_OUT qq{\t<deletedfiles count="}.(keys %{$self->{_DELETEDFILES}}).qq{">\n});
		foreach (keys %{$self->{_DELETEDFILES}}) {
			print(F_OUT qq{\t\t<deletedfile><![CDATA[$_]]></deletedfile>\n});
		}		
		print(F_OUT qq{\t</deletedfiles>\n});		
		
		print(F_OUT qq{</changes>\n});		
		close(F_OUT);		
	}	
	
	sub filecheck::_log($$$) {
		# Any errors are written to private member $self->{_LOG}
		my $self = shift();
		my $file = shift();
		my $error = shift();
		
		$self->{_LOG}->{$file} = $error;
		print("$file: ".$self->{_LOG}->{$file}."\n");
		next();
	}
	
	sub filecheck::_writeerrorlog($) {
		# Private member $self->{_LOG}, if it is !undef is written to an
		# xml file
		my $self = shift();
		
		open(F_OUT, ">", $self->{_CONFIGURATION}->errors()) || die($!);
		print(F_OUT qq{<?xml version="1.0" encoding="iso-8859-1"?>\n});
		print(F_OUT qq{<errors count="}.(keys %{$self->{_LOG}}).qq{" created="}.localtime(time()).qq{">\n});

		foreach (keys %{$self->{_LOG}}) {
			print(F_OUT qq{\t<error file="$_" error="$self->{_LOG}->{$_}" />\n});
		}
		
		print(F_OUT "<\/errors>\n");	
		close(F_OUT);			
	}

1;
